import json
import socket

from .files import get_file_contents


def read_description(sd_filename):
    sd = get_file_contents(sd_filename)
    my_ip = find_my_ip()
    if my_ip:
        sd = sd.replace("localhost", my_ip)
    return json.loads(sd)


def find_my_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except:
        ip = None
    finally:
        s.close()
    return ip



import logging
import urllib.parse
from functools import wraps
from urllib.parse import urlparse
import sys, signal

from .broker_client import BrokerClient
from .agent import Agent
from .service import find_my_ip, read_description
from .access_control import read_policies, build_access_request

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

class Provider(Agent):
    def __init__(self, **kwargs):
        logging.info("Initializing provider with args: %s" % kwargs)
        super().__init__(**kwargs)
        self.set_exit_trap()
        if "description_file" in kwargs.keys():
            self.description_file = kwargs["description_file"]
            self.description = read_description(self.description_file)
            self.update_ip()

        if "policies_file" in kwargs.keys():
            self.policies_file = kwargs["policies_file"]
            self.policies = read_policies(self.description, self.policies_file)

    def update_ip(self):
        my_ip = find_my_ip()
        self.description["@id"] = self.replace_ip(self.description["@id"], my_ip)
        self.description["broker"] = self.replace_ip(self.description["broker"], my_ip)

    def replace_ip(self, url, ip):
        parsed_url = urlparse(url)
        parsed_url = parsed_url._replace(netloc="%s:%s" % (ip, parsed_url.port))
        return parsed_url.geturl()

    def join_swarm(self):
        return self.register() and self.setup_policies()

    def register(self):
        return BrokerClient.register(self)

    def setup_policies(self):
        return BrokerClient.setup_policies(self.description, self.policies)

    def enforce_authorization(self, f):
        @wraps(f)
        def wrap(*args, **kwargs):
            from flask import request, jsonify
            access_request = build_access_request(self.description, request.headers, request.method)
            allowed, error = BrokerClient.authorize(self.description, access_request)
            if allowed:
                return f(*args, **kwargs)
            else:
                return jsonify(error)
        return wrap

    def port(self):
        return urlparse(self.description["@id"]).port

    def handle_exit(self, sig, frame):
        BrokerClient.unregister(self.description)
        sys.exit(0)

    def set_exit_trap(self):
        signal.signal(signal.SIGINT, self.handle_exit)

    def __repr__(self):
        rep = super().__repr__()
        rep["policies"] = list(map(lambda p: p["name"], self.policies))
        return rep



import urllib
import json
import uuid
import requests

import logging
logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

from .files import *


def read_policies(description, filename):
    if not exists(filename):
        write_file_contents(json.dumps([]), filename)
        return []

    policies = get_file_contents(filename)
    policies = policies.replace("@type", description["@type"])
    policies_json = json.loads(policies)
    policies_json = add_id_to_policies(policies_json)
    write_file_contents(json.dumps(policies_json, indent=4), filename)
    return policies_json

def add_id_to_policies(policies):
    for policy in policies:
        if not "id" in policy:
            policy["id"] = str(uuid.uuid4())

    return policies

def gather_user_attrs_from_headers(headers):
    if "x-user-attrs" in headers:
        # user attributes header is encoded as 'application/x-www-form-urlencoded'
        x_user_attrs = urllib.parse.parse_qs(headers["x-user-attrs"])
    else:
        x_user_attrs = {}

    return {name:value.pop() for name, value in x_user_attrs.items()}

def gather_object_attrs(description):
    print("====== " + description["@id"])
    return {"swarm:ServiceType": description["@type"], "swarm:Id": description["@id"]}

def build_access_request(description, headers, method):
    user_attrs = gather_user_attrs_from_headers(headers)
    object_attrs = gather_object_attrs(description)
    operations = [restful_methods_to_operations()[method]]

    return {
        "user_attrs": user_attrs,
        "object_attrs": object_attrs,
        "context_attrs": {},
        "operations": operations,
    }

def restful_methods_to_operations():
    return {
        "GET": "read",
        "POST": "create",
        "PUT": "update",
        "DELETE": "delete",
    }

def operations_to_restful_methods():
    return {
        v: k for k, v in restful_methods_to_operations().items()
    }



import base58
import json
import os
from Crypto.Hash import SHA256
from hashlib import sha256
from ecdsa import SigningKey, VerifyingKey, SECP256k1

DEFAULT_FILE_NAME = 'keys.json'

def generate_keys_and_wallet_id(filename=DEFAULT_FILE_NAME):
    sk = SigningKey.generate(curve=SECP256k1, hashfunc=sha256)
    vk = sk.get_verifying_key()
    signature = sk.sign(b'message')
    assert vk.verify(signature, b'message')
    pub_key_bin = vk.to_string()
    wallet_id = get_wallet_id(pub_key_bin)
    private_key = base58.b58encode(sk.to_string())
    public_key = base58.b58encode(pub_key_bin)
    keys_dict = {'private-key':private_key.decode(),'public-key':public_key.decode(),'walletId':wallet_id}
    with open(filename, 'w') as file_obj:
        file_obj.write(json.dumps(keys_dict))
    return keys_dict

def get_keys_and_wallet_id(filename=None):
    if filename:
        if os.path.exists(str(filename)):
            with open(filename, 'r') as file_obj:
                keys_dict = json.loads(file_obj.read())
                keys_dict['private-key'] = base58.b58decode(keys_dict['private-key'].encode())
                keys_dict['public-key'] = base58.b58decode(keys_dict['public-key'].encode())
                return keys_dict
        else:
            keys_dict = generate_keys_and_wallet_id(filename)
            keys_dict['private-key'] = base58.b58decode(keys_dict['private-key'].encode())
            keys_dict['public-key'] = base58.b58decode(keys_dict['public-key'].encode())
            return keys_dict
    elif os.path.exists(DEFAULT_FILE_NAME):
        with open(DEFAULT_FILE_NAME, 'r') as file_obj:
            keys_dict = json.loads(file_obj.read())
            keys_dict['private-key'] = base58.b58decode(keys_dict['private-key'].encode())
            keys_dict['public-key'] = base58.b58decode(keys_dict['public-key'].encode())
            return keys_dict
    else:
        keys_dict = generate_keys_and_wallet_id()
        keys_dict['private-key'] = base58.b58decode(keys_dict['private-key'].encode())
        keys_dict['public-key'] = base58.b58decode(keys_dict['public-key'].encode())
        return keys_dict

def get_wallet_id(pub_key_bin):
    pub_hash_bin = SHA256.new(pub_key_bin).digest()
    wallet_id = base58.b58encode_check(pub_hash_bin)
    return wallet_id.decode()

def get_msg_to_check_sig(data):
    data = [data['hash'], data['n']]
    return json.dumps(data, separators=(',', ':'))

def sign_tx(transaction, keys_dict):
    if type(transaction) == dict:
        for input in transaction['content']['input']:
            prev_out = input['prev_out']
            msg_signed_bin = get_msg_to_check_sig(prev_out).encode()
            sk = SigningKey.from_string(keys_dict['private-key'], curve=SECP256k1, hashfunc=sha256)
            signature = sk.sign(msg_signed_bin)
            vk = VerifyingKey.from_string(keys_dict['public-key'], curve=SECP256k1, hashfunc=sha256)
            assert vk.verify(signature, msg_signed_bin)
            sig = base58.b58encode_check(signature).decode()
            puk = base58.b58encode_check(keys_dict['public-key']).decode()
            input['sig'] = sig
            input['pub_key'] = puk
        return transaction
    else:
        return None



# -*- coding: utf-8 -*-

from .broker_client import *
from .agent import Agent
from .consumer import Consumer
from .provider import Provider

from .files import *
from .economy import *
from .service import read_description, find_my_ip
from .access_control import read_policies, build_access_request, operations_to_restful_methods



import logging
from .agent import Agent
from .service import find_my_ip

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

class Consumer(Agent):
    def __init__(self, **kwargs):
        logging.info("Initializing consumer with args: %s" % kwargs)
        super().__init__(**kwargs)
        self.description["@type"] = "swarm:Consumer"
        self.description["@id"] = find_my_ip()
        self.description["broker"] = "http://localhost:4000/broker"
        self.description["pricing"] = {"walletId": self.keys["walletId"]}

    def __repr__(self):
        return super().__repr__()



import os

def exists(filename):
	return os.path.isfile(filename)

def write_file_contents(data, filename="default-policies.json"):
    sd_file = open(filename, "w")
    sd_file.write(data)
    sd_file.close()

def get_file_contents(filename="service-description.jsonld"):
    sd_file = open(filename, "r")
    sd = sd_file.read()
    sd_file.close()
    return sd



import logging
from .broker_client import BrokerClient

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

class ServiceExecutor():
    def __init__(self, agent, query):
        logging.info("Initializing consumer with query: %s" % query)
        self.state = "initialized"
        self.agent = agent
        self.query = query
        self.contract = None
        self.contract_location = None

    def contract_providers(self):
        if self.search_and_contract() and self.accept_negotiation():
            logging.info("Contract worked")
            return self.contract
        else:
            logging.info("Contract failed at state %s" % self.state)
            return None

    def search_and_contract(self):
        negotiation = BrokerClient.contract(self.agent.description, {"query": self.query})
        if not negotiation:
            self.state = "search_and_contract_failed"
            return None
        else:
            self.state = "waiting_accept_negotiation"
            self.contract, self.contract_location = negotiation
            return self.contract

    def accept_negotiation(self):
        contract = BrokerClient.sign_and_send_tx(self.agent.description, self.agent.keys, self.contract, self.contract_location)
        if not contract:
            self.state = "accept_negotiation_failed"
            return None
        else:
            self.state = "sla_established"
            self.contract = contract
            return contract

    def execute(self, *args, **kwargs):
        if self.state == "sla_established":
            return BrokerClient.smart_use(self.contract, *args, **kwargs)
        else:
            logging.debug("Use service: invalid state '%s'" % self.state)

    def decode(self, responses):
        return list(map(lambda r: r.json(), responses))

    def combine_feedbacks(self, broker_feedback, party_feedback, alfa=0.5):
        return alfa * party_feedback + (1 - alfa) * broker_feedback

    def build_reputation_output(self, agreement):
        consumer_reputation = 2.5 # FIXME, e.g., ReputationClientMock.lookup(self.agent.description["@id"])
        broker_feedback = agreement["broker_feedback"]
        consumer_feedback = 2.5 # FIXME
        return {
            "service_id": agreement["provider_id"],
            "broker_feedback": broker_feedback,
            "party_feedback": consumer_feedback,
            "reputation": self.combine_feedbacks(broker_feedback, consumer_feedback, consumer_reputation/5),
        }

    def compute_and_send_reputation(self):
        reputation_tx = {
            "hash_payment_tx": self.contract["tx"]["hash"],
            "party_id": self.agent.description["@id"],
            "party_type": "consumer",
            "output": list(map(self.build_reputation_output, self.contract["sla"])),
            "input": None
        }
        reputation_tx["input"] = {
            "hash_last_tx": "get_last_tx_hash(self.agent.description['@id'])", # FIXME NOW
            "party_sig": "sign_reputation_tx(reputation_tx)", # FIXME NOW
            "party_pub_key": "get_public_key(self.agent.description['@id'])" # FIXME NOW
        }
        logging.debug("Created reputation_tx: %s" % reputation_tx)
        return BrokerClient.send_reputation(self.agent.description, reputation_tx, self.contract_location)

    def get_num_services(self):
        if self.contract:
            return len(self.contract['tx']['content']['output']) - 1
        else:
            logging.debug("get_num_services: contract is None")

    def contract_summary(self):
        if self.contract:
            summary = {
                "consumer_balance": self.contract["tx"]["content"]["output"][0]["value"],
                "providers": [{"provider": o["provider_id"], "price": o["value"]} for o in self.contract["tx"]["content"]["output"][1:]]
            }
            if "sig" in self.contract["tx"]["content"]["input"][0].keys():
                summary["signature"] = self.contract["tx"]["content"]["input"][0]["sig"]

            return summary



import json
import logging

from .broker_client import BrokerClient
from .service_executor import ServiceExecutor
from .economy import get_keys_and_wallet_id

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

class Agent():
    """Superclass for instantiating a Swarm service"""
    def __init__(self, **kwargs):
        self.description = self.new_description()
        if "keys_file" in kwargs.keys():
            self.keys_file = kwargs["keys_file"]
            self.keys = get_keys_and_wallet_id(self.keys_file)
            self.update_wallet_id(self.keys["walletId"])

    def new_description(self):
        return {"pricing": {}, "operations": [], "geolocation": {}}

    def setup_query(self, query):
        query["metadata"] = {
            "requester": {
                "@type": self.description["@type"],
                "@id": self.description["@id"],
                "pricing": {"walletId": self.description["pricing"]["walletId"]}
            }
        }
        return query

    def update_wallet_id(self, walletId):
        self.description["pricing"]["walletId"] = self.keys["walletId"]

    def discover(self, query):
        query = self.setup_query(query)
        return BrokerClient.discover(self.description, query)

    def get_executor(self, query):
        query = self.setup_query(query)
        return ServiceExecutor(self, query)

    def __repr__(self):
        return {
            "description": {
                "@id": self.description["@id"],
                "@type": self.description["@type"],
                "broker": self.description["broker"],
                "walletId": self.description["pricing"]["walletId"]
            }
        }

    def __str__(self):
        return str(self.__repr__())



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='swarm_lib',
    version='0.0.11',
    description='A Library for accessing the Swarm network written in Python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Geovane Fedrecheski',
    install_requires=[
        "requests",
        "base58",
        "ecdsa",
        "flask",
        "wrap",
        "pycrypto"
    ],
    packages=find_packages(exclude=('tests', 'docs'))
)

print(find_packages())



# -*- coding: utf-8 -*-

import requests
import json
import logging
import time
from threading import Timer
from functools import wraps
import urllib.parse
import re

from .files import *
from .economy import *

from .service import read_description, find_my_ip
from .access_control import read_policies, build_access_request, operations_to_restful_methods

logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

class BrokerClient():

    # provider functions

    def register(provider, retries=0, max_retries=10):
        registry_url = provider.description["broker"] + "/api/v2/registry"
        logging.info("Will register service %s at %s" % (provider.description["@id"], registry_url))
        try:
            response = requests.post(registry_url, json=provider.description)
            if response.status_code == 200:
                logging.info("Registered service %s" % (provider.description["@type"]))
                return True
        except Exception as e:
            if retries < max_retries:
                retries += 1
                sleep_time = 2 ** retries
                logging.error("Registry failed: no broker found. Will retry (%s) in %s seconds" % (retries, sleep_time))
                time.sleep(sleep_time)
                provider.update_ip()
                return BrokerClient.register(provider, retries)
            else:
                logging.error("No available broker at %s to register at" % registry_url)
                return False

    def setup_policies(service_desc, policies):
        policies_url = service_desc["broker"] + "/api/v2/security/policies"
        try:
            response = requests.post(policies_url, json=policies)
            policy_names = list(map(lambda p: p["name"], policies))
            if response.status_code == 200:
                logging.info("Added policies %s" % (policy_names))
                return True
            else:
                logging.error("Error adding policies %s. Status = %d" % (policy_names, response.status_code))
                return False
        except Exception as e:
            logging.debug(e)
            logging.error("No available broker at %s to add policies" % policies_url)
            return False

    def authorize(service_desc, access_request):
        pdp_url = service_desc["broker"] + "/api/v2/security/authorizations"
        try:
            result = requests.post(pdp_url, json=access_request)
            if result.status_code == 200:
                logging.debug("Request was authorized: %s" % access_request)
                return True, None
            else:
                logging.debug("Request was denied: %s" % access_request)
                return False, {"error": "access denied", "user_attrs": access_request["user_attrs"]}
        except Exception as e:
            logging.error(e)
            return False, {"error": str(e), "user_attrs": access_request["user_attrs"]}

    # provider and consumer functions

    def discover(service_desc, query={"@type": "*"}):
        discovery_url = service_desc["broker"] + "/api/v2/discovery-requests"
        try:
            logging.info("Discovery query %s" % query)
            response = requests.post(discovery_url, json=query)
            response_json = response.json()
            logging.info("Discovered services are %s" % list(map(lambda c: c["description"]["@id"], response_json["candidates"])))
            return list(map(lambda c: c["description"], response_json["candidates"]))
        except Exception as e:
            logging.debug("The discovery request at %s did not work" % discovery_url)
            logging.debug(e)
            return []

    def contract(service_desc, data={"query": {"@type": "*"}}):
        contract_url = service_desc["broker"] + "/api/v2/contracts"
        try:
            logging.info("Running contract for %s" % data)
            response = requests.post(contract_url, json=data)
            location = response.headers['location']
            contract = response.json()
            logging.info("State dict created is %s '\n' And the transaction signed\
                         must be sent to location %s" % (json.dumps(contract),location))
            return (contract, location)
        except Exception as e:
            logging.debug("The contract request at %s did not work" % contract_url)
            return None

    def sign_and_send_tx(service_desc, keys, contract, location):
        signed_tx = sign_tx(contract['tx'], keys)
        logging.debug("Signed tx: %s" % (signed_tx))
        url = service_desc["broker"] + location
        if signed_tx and location:
            try:
                response = requests.put(url, json={"tx": signed_tx})
                if response.status_code == 200:
                    contract = BrokerClient.wait_sla_establishment(url)
                    logging.info("The negotiation is complete")
                    return contract
            except Exception as e:
                logging.debug(e)
                logging.info("Sending the signed transaction did not work")
                return None
        else:
            logging.error("signed_tx: %s\nlocation: %s" % (signed_tx, location))
            return None

    def wait_sla_establishment(url):
        status, max_retries, retries = 0, 3, 0
        block_period = 3
        while status != 200 and retries < max_retries:
            response = requests.get(url)
            status = response.status_code
            retries += 1
            if status != 200:
                time.sleep(block_period)
        if retries <= max_retries:
            return response.json()
        else:
            return None

    def smart_use(contract, data_array=None):
        if not data_array:
            data_array = [None for sla in contract["sla"]]

        responses, sla_index = [], 0
        for data in data_array:
            sla = contract["sla"][sla_index]
            sla_index = (sla_index + 1) % len(contract["sla"])

            response = BrokerClient.smart_use_one(contract["tx"], sla, data)
            if response:
                responses.append(response)

        return responses

    def smart_use_one(tx, sla, data):
        token = sla["delegation_token"]
        method, url = BrokerClient.http_bind(tx, sla)

        user_attrs = {"swarm:Id": tx["content"]["consumer_id"]}

        if method:
            return BrokerClient.use(method, url, data, token, user_attrs)
        else:
            logging.warn("Could not infer method (%s)" % (method))
            return False

    def http_bind(tx, sla):
        provider_description = sla["provider"]
        operation = tx["content"]["operations"]
        method = operations_to_restful_methods()[operation]
        url = provider_description["@id"] + provider_description["operations"][0]["entry"]
        return method, url

    def use(method, url, data=None, token=None, user_attrs={}):
        if token:
            user_attrs["swarm:Token"] = token
        headers = {'x-user-attrs': urllib.parse.urlencode(user_attrs)}

        logging.debug([method, url, headers, data])
        response = requests.request(method, url, headers=headers, json=data)
        logging.debug("Request to %s is %s" % (url, response.status_code))
        return response

    def send_reputation(service_desc, reputation_tx, location):
        if not location:
            logging.error("Location cannot be empty")
            return False

        url = service_desc["broker"] + location + "/reputation"
        logging.debug("Will send reputation_tx %s to %s" % (reputation_tx, url))

        try:
            response = requests.post(url, json={"reputation_tx": reputation_tx})
            if response.status_code == 200:
                logging.info("The reputation_tx was sent")
                return response.json()
        except Exception as e:
            logging.debug(e)
            logging.info("Sending reputation_tx did not work")
            return None

    def unregister(service_desc):
        quoted_id = urllib.parse.quote(service_desc["@id"], safe="")
        registry_url = service_desc["broker"] + "/api/v2/registry/" + quoted_id
        response = requests.delete(registry_url)
        if response.status_code == 200:
            logging.info("De-registered service %s" % (service_desc["@id"]))
